import logging
from dataclasses import dataclass
from functools import cached_property
from io import BytesIO
from pathlib import Path
from typing import Optional

import fitz
import pdftotext
from pdf2john import PdfHashExtractor as EncryptionMetadataExtractor

from monopoly.config import PdfConfig

logger = logging.getLogger(__name__)


@dataclass
class PdfPage:
    """
    Dataclass representation of a bank statement PDF page.
    Contains the raw text of a PDF page, and allows access
    to the raw text as a list via the `lines` property
    """

    raw_text: str

    @cached_property
    def lines(self) -> list[str]:
        return self.raw_text.split("\n")


class PdfParser:
    def __init__(
        self,
        file_path: Path,
        pdf_config: Optional[PdfConfig] = None,
    ):
        """
        Class responsible for parsing PDFs and returning raw text

        The page_range variable determines which pages are extracted.
        All pages are extracted by default.
        """
        self.file_path = file_path

        if pdf_config is None:
            pdf_config = PdfConfig()

        self.passwords = pdf_config.passwords
        self.page_range = slice(*pdf_config.page_range)
        self.page_bbox = pdf_config.page_bbox

    def open(self):
        """
        Opens and decrypts a PDF document
        """
        logger.debug("Opening pdf from path %s", self.file_path)
        document = self.document

        if not document.is_encrypted:
            return document

        if self.passwords:
            for password in self.passwords:
                document.authenticate(password)

                if not document.is_encrypted:
                    logger.debug("Successfully authenticated with password")
                    return document

        raise ValueError(f"Wrong password - unable to open document {document}")

    def get_pages(self) -> list[PdfPage]:
        logger.debug("Extracting text from PDF")
        document: fitz.Document = self.open()

        num_pages = list(range(document.page_count))
        document.select(num_pages[self.page_range])

        for page in document:
            if self.page_bbox:
                logger.debug("Cropping page")
                page.set_cropbox(self.page_bbox)

            logger.debug("Removing vertical text")
            page = self._remove_vertical_text(page)

        # certain statements require garbage collection, so that duplicate objects
        # do not cause pdftotext to fail due to missing xrefs/null values
        # however, setting `garbage=2` may cause issues with other statements
        # so an initial attempt should be made to run using `garbage=0`
        garbage_values = [0, 2]

        for garbage in garbage_values:
            try:
                pdf_byte_stream = BytesIO(document.tobytes(garbage=garbage))
                pdf = pdftotext.PDF(pdf_byte_stream, physical=True)
                return [PdfPage(page) for page in pdf]
            except pdftotext.Error:
                continue
        raise RuntimeError("Unable to retrieve pages")

    @cached_property
    def document(self) -> fitz.Document:
        """
        Returns a Python representation of a PDF document.
        """
        return fitz.Document(self.file_path)

    @cached_property
    def extractor(self) -> EncryptionMetadataExtractor:
        """
        Returns an instance of pdf2john, used to retrieve and return
        the encryption metadata from a PDF's encryption dictionary
        """
        return EncryptionMetadataExtractor(self.file_path)

    @staticmethod
    def _remove_vertical_text(page: fitz.Page):
        """Helper function to remove vertical text, based on writing direction (wdir).

        This helps avoid situations where the PDF is oddly parsed, due to vertical text
        inside the PDF.

        An example of vertical text breaking a transaction:
        ```
        'HEALTHY HARVEST CAFÉ SINGAPORE SG',
        'Co Reg No: 123456',
        '10 NOV 9.80',
        ```

        Note:
            The 'dir' key represents the tuple (cosine, sine) for the angle.
            If line["dir"] != (1, 0), the text of its spans is rotated.

        """
        for block in page.get_text("dict", flags=fitz.TEXTFLAGS_TEXT)["blocks"]:
            for line in block["lines"]:
                writing_direction = line["dir"]
                if writing_direction != (1, 0):
                    page.add_redact_annot(line["bbox"])
        page.apply_redactions(images=fitz.PDF_REDACT_IMAGE_NONE)
        return page



import logging

from monopoly.config import CreditStatementConfig, DebitStatementConfig
from monopoly.constants import (
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EncryptionIdentifier,
    MetadataIdentifier,
    StatementBalancePatterns,
)

from ..base import ProcessorBase

logger = logging.getLogger(__name__)


class Dbs(ProcessorBase):
    credit_config = CreditStatementConfig(
        bank_name=BankNames.DBS,
        statement_date_pattern=r"(\d{2}\s[A-Za-z]{3}\s\d{4})",
        statement_date_format=r"%d %b %Y",
        transaction_date_format=r"%d %b",
        multiline_transactions=True,
        transaction_pattern=CreditTransactionPatterns.DBS,
        prev_balance_pattern=StatementBalancePatterns.DBS,
    )

    debit_config = DebitStatementConfig(
        bank_name=BankNames.DBS,
        statement_date_pattern=r"(\d{2}\s[A-Za-z]{3}\s\d{4})",
        statement_date_format=r"%d %b %Y",
        transaction_date_format=r"%d %b",
        multiline_transactions=True,
        debit_statement_identifier=r"(WITHDRAWAL.*DEPOSIT.*BALANCE)",
        transaction_pattern=DebitTransactionPatterns.DBS,
    )

    identifiers = [
        EncryptionIdentifier(
            pdf_version=1.4, algorithm=2, revision=3, length=128, permissions=-1852
        ),
        MetadataIdentifier(creator="Quadient CXM AG"),
    ]



import logging
import re
from functools import cached_property

import fitz

from monopoly.config import DebitStatementConfig
from monopoly.constants import AccountType, StatementFields
from monopoly.pdf import PdfPage

from .base import BaseStatement, Transaction

logger = logging.getLogger(__name__)


class DebitStatement(BaseStatement):
    """
    A dataclass representation of a debit statement
    """

    def __init__(
        self,
        document: fitz.Document,
        pages: list[PdfPage],
        debit_config: DebitStatementConfig,
    ):
        self.config = debit_config
        self.statement_type = AccountType.DEBIT
        super().__init__(document, pages, debit_config)

    def _process_line(
        self,
        line: str,
        lines: list[str],
        idx: int,
        pattern: re.Pattern,
    ) -> Transaction | None:
        if match := pattern.search(line):
            groupdict = match.groupdict()
            suffix = self.get_debit_suffix(line, pattern)

            # if the entry is a credit then add parenthesis around
            # the amount, which causes the Transaction class to
            # treat the amount as a credit / 'positive' transaction
            if suffix == "CR":
                line = line.replace(groupdict["amount"], f"({groupdict['amount']})")
                lines[idx] = line

            # continue with normal line processing
            return super()._process_line(line, lines, idx, pattern)
        return None

    def get_debit_suffix(self, line: str, pattern: re.Pattern):
        """
        Gets the accounting suffix for debit card statements

        This is necessary, since the amount in the row does not have any
        identifier apart from column position.
        """
        if match := pattern.search(line):
            amount = match[StatementFields.AMOUNT]
            pos = line.find(amount)
            withdrawal_diff = abs(pos - self.withdrawal_pos)
            deposit_diff = abs(pos - self.deposit_pos)
            if withdrawal_diff > deposit_diff:
                return "CR"
            return "DR"
        return None

    @cached_property
    def debit_header(self) -> str | None:
        if self.config and self.config.debit_statement_identifier:
            for line in self.pages[0].lines:
                if re.search(self.config.debit_statement_identifier, line):
                    return line.lower()
        return None

    @cached_property
    def withdrawal_pos(self) -> int:
        if self.debit_header:
            return self.debit_header.find("withdrawal")
        raise ValueError("Debit header missing")

    @cached_property
    def deposit_pos(self) -> int:
        if self.debit_header:
            return self.debit_header.find("deposit")
        raise ValueError("Debit header missing")

    def perform_safety_check(self) -> bool:
        """
        Checks that debit and credit transaction sums
        exist as a number within the statement
        """
        df = self.df
        amount = StatementFields.AMOUNT

        numbers = self.get_all_numbers_from_document()

        # add a zero, for cases where the debit statement
        # either is completely debit or credit transactions
        numbers.update([0])

        debit_sum = round(abs(df[df[amount] > 0][amount].sum()), 2)
        credit_sum = round(abs(df[df[amount] < 0][amount].sum()), 2)

        result = all([debit_sum in numbers, credit_sum in numbers])

        if not result:
            logger.warning(self.warning_message)

        return result



from glob import glob

from monopoly.processors import Ocbc


def ocbc_example():
    """Example showing how monopoly can be used to extract data from
    multiple bank statements in a directory
    """
    ocbc_statements = glob("statements/ocbc/*.pdf")
    for file_path in ocbc_statements:
        processor = Ocbc(
            file_path=file_path,
        )
        statement = processor.extract()
        transformed_df = processor.transform(statement)
        processor.load(
            transformed_df, statement, output_directory="src/monopoly/examples"
        )


if __name__ == "__main__":
    ocbc_example()



import logging
from datetime import datetime
from pathlib import Path

from pandas import DataFrame

from monopoly.config import CreditStatementConfig, DebitStatementConfig
from monopoly.constants import StatementFields
from monopoly.pdf import PdfParser
from monopoly.statements import CreditStatement, DebitStatement
from monopoly.write import generate_name

logger = logging.getLogger(__name__)


class StatementProcessor:
    """
    Handles extract, transform and load (ETL) logic for bank statements

    Transactions are extracted from pages, before undergoing various
    transformations like converting various date formats into ISO 8601.
    Transformed statements are then written to a CSV file.
    """

    credit_config: CreditStatementConfig
    debit_config: DebitStatementConfig
    statement: CreditStatement | DebitStatement

    def __init__(
        self,
        file_path: Path,
        parser: PdfParser,
        statement: CreditStatement | DebitStatement,
    ):
        self.file_path = file_path
        self.parser = parser
        self.statement = statement

    def extract(self) -> CreditStatement | DebitStatement:
        """Extracts transactions from the statement, and performs
        a safety check to make sure that total transactions add up"""
        statement = self.statement

        if not statement.transactions:
            raise ValueError("No transactions found - statement extraction failed")

        if not statement.statement_date:
            raise ValueError("No statement date found")

        statement.perform_safety_check()

        return statement

    def transform(self, statement: CreditStatement | DebitStatement) -> DataFrame:
        logger.debug("Running transformation functions on DataFrame")

        df = statement.df
        statement_date = statement.statement_date
        date_format = statement.statement_config.transaction_date_format

        def convert_date(row):
            """
            Converts each date to a ISO 8601 (YYYY-MM-DD) format.

            Implements cross-year logic by attributing transactions from
            October, November, and December to the previous year if
            the statement month is January.
            e.g. if the statement month is Jan 2024, transactions from
            Oct/Nov/Dec should be attributed to the previous year.
            """
            parsed_date = datetime.strptime(
                row[StatementFields.TRANSACTION_DATE], date_format
            )
            row_year = statement_date.year
            row_day, row_month = parsed_date.day, parsed_date.month

            if statement_date.month == 1 and row_month != 1:
                row_year = statement_date.year - 1

            return f"{row_year}-{row_month:02d}-{row_day:02d}"

        logger.debug("Transforming dates to ISO 8601")
        df[StatementFields.TRANSACTION_DATE] = df.apply(convert_date, axis=1)
        return df

    def load(
        self,
        df: DataFrame,
        statement: CreditStatement | DebitStatement,
        output_directory: Path,
    ):
        if isinstance(output_directory, str):
            output_directory = Path(output_directory)

        filename = generate_name(
            file_path=self.file_path,
            format_type="file",
            statement_config=statement.statement_config,
            statement_type=statement.statement_type,
            statement_date=statement.statement_date,
        )

        output_path = output_directory / filename
        logger.debug("Writing CSV to file path: %s", output_path)
        df.to_csv(output_path, index=False)

        return output_path



import logging
import sys


def get_logger() -> logging.Logger:
    logger = logging.getLogger("root")
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    logging.getLogger("pdf2john").setLevel(logging.ERROR)
    logging.getLogger("pyhanko").setLevel(logging.ERROR)
    return logger



import re
from functools import cached_property
from pathlib import Path
from typing import Any, Optional

from monopoly.config import CreditStatementConfig, DebitStatementConfig, PdfConfig
from monopoly.constants import EncryptionIdentifier, MetadataIdentifier
from monopoly.pdf import PdfParser
from monopoly.processor import StatementProcessor
from monopoly.statements import CreditStatement, DebitStatement


class ProcessorBase(StatementProcessor):
    """Helper class to handle initialization of common variables
    that are shared between bank processor classes

    Also, helps to parse the statement, and create a Credit or Debit
    Statement class object depending on whether a regex match is found
    for the debit_statement_identifier variable
    """

    # overwritten if pdf_config exists in bank class
    pdf_config: PdfConfig = PdfConfig()
    identifiers: list[EncryptionIdentifier | MetadataIdentifier]

    def __init__(
        self,
        file_path: Path,
        passwords: Optional[list[str]] = None,
    ):
        # this allows the user to override the default pydantic password
        # and supply their own password via the bank subclasses
        if passwords:
            self.pdf_config.passwords = passwords

        parser = PdfParser(file_path, self.pdf_config)

        self.pages = parser.get_pages()
        self.document = parser.document

        if isinstance(file_path, str):
            file_path = Path(file_path)

        super().__init__(
            parser=parser, file_path=file_path, statement=self.get_statement()
        )

    @cached_property
    def statement_config(self) -> CreditStatementConfig | DebitStatementConfig:
        if self.debit_header and self.debit_config:
            return self.debit_config
        return self.credit_config

    def get_statement(self) -> CreditStatement | DebitStatement:
        if not hasattr(self, "debit_config"):
            return CreditStatement(self.document, self.pages, self.credit_config)
        if self.debit_header and self.debit_config:
            return DebitStatement(self.document, self.pages, self.debit_config)
        return CreditStatement(self.document, self.pages, self.credit_config)

    @cached_property
    def debit_header(self) -> str | None:
        """Returns the 'header' line of the debit statement

        Used to determine whether a statement is a debit statement, and also
        determine whether transactions in a debit statement should be treated as a
        debit or credit entry
        """
        if self.debit_config and self.debit_config.debit_statement_identifier:
            for line in self.pages[0].lines:
                if re.search(self.debit_config.debit_statement_identifier, line):
                    return line.lower()
        return None

    @staticmethod
    def get_identifiers(parser: PdfParser) -> list[Any]:
        """Retrives encryption and metadata identifiers from a bank statement PDF

        Used to automatically select the correct bank processing class
        """
        identifiers = []
        # pylint: disable=protected-access
        if parser.extractor.encrypt_dict:
            extractor = parser.extractor
            major, minor = extractor.pdf._header_version
            pdf_version = f"{major}.{minor}"
            encryption_identifier = EncryptionIdentifier(
                float(pdf_version),
                extractor.algorithm,
                extractor.revision,
                extractor.length,
                extractor.permissions,
            )
            identifiers.append(encryption_identifier)

        if metadata := parser.document.metadata:
            metadata_identifier = MetadataIdentifier(**metadata)
            identifiers.append(metadata_identifier)  # type: ignore

        if not identifiers:
            raise ValueError(f"Could not get identifier for {parser.file_path}")

        return identifiers
